part of disposable;

/// Represents a group of disposable resources that are disposed together.
class CompositeDisposable implements Disposable {
  final List<Disposable> _disposables;

  bool _isDisposed = false;

  @override
  bool get isDisposed => _isDisposed;

  /// Returns the number of disposables contained in the [CompositeDisposable].
  int get length => _disposables.length;

  /// Initializes a new instance of the [CompositeDisposable] class
  /// with no disposables contained by it initially.
  CompositeDisposable()
      : _disposables = new List<Disposable>();

  /// Initializes a new instance of the [CompositeDisposable] class
  /// from a group of disposables.
  CompositeDisposable.from(Iterable<Disposable> disposables)
      : _disposables = new List<Disposable>.from(disposables);

  /// Adds a [disposable] to the [CompositeDisposable] or disposes
  /// the [disposable] if the [CompositeDisposable] is disposed.
  void add(Disposable disposable) {
    if (disposable == null) throw new ArgumentError.notNull('disposable');
    if (_isDisposed) disposable.dispose();
    _disposables.add(disposable);
  }

  /// Removes and disposes the first occurrence of a [disposable]
  /// from the [CompositeDisposable].
  void remove(Disposable disposable) {
    if (disposable == null) throw new ArgumentError.notNull('disposable');
    if (!_disposables.contains(disposable)) return;

    int index = _disposables.indexOf(disposable);
    _disposables[index].dispose();
    _disposables.removeAt(index);
  }

  /// Determines whether the [CompositeDisposable] contains a specific
  /// [disposable].
  bool contains(Disposable disposable) {
    if (disposable == null) throw new ArgumentError.notNull('disposable');
    if (_isDisposed) return false;

    return _disposables.contains(disposable);
  }

  /// Removes and disposes all disposables from the CompositeDisposable,
  /// but does not dispose the [CompositeDisposable].
  void clear() {
    var errors = new List();
    _disposables.forEach((_) {
      try {
        _.dispose();
      } catch (ex) {
        errors.add({'target': _, 'error': ex});
      }
    });
    // Failed items are referenced in the error so the list can be cleared
    _disposables.clear();
    if (errors.length > 0) throw errors;
  }

  /// Disposes all disposables in the group and removes them from the group.
  @override
  void dispose() {
    if (_isDisposed) return;

    _isDisposed = true;
    clear();
  }
}
